function [imgH] = LLKR(imgL, dictH, dictL, M, C, idxM, overlap, ...
    scale, gradThresh, numNN)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Testing Phase of Local Learnable Kernel Regression (LLKR)
%
% Input :
%       imgL : (h x w x 3) input LR image
%      dictH : (dH x N) dictionary of HR patch
%      dictL : (dL x N) dictionary of LR patch (Note: multiplied by A)
%          M : (k x 1 cell array) k metric matrix A
%          C : (dH x K) clustering centers
%       idxM : (k x 1 cell array) index of HR patch 
%    overlap : overlap length of neighbor patches
%      scale : factor of magnification
% gradThresh : threshold of gradient
%      numNN : number of nearest neighbours for LLKR
% 
% Output
%       imgH : output HR image
%
% Author
%   Renjie Liao
% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

patchSize = sqrt(size(dictH, 1));
numMetrics = size(M, 1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% read input LR image
if 3 == ndims(imgL)
    imgLYcbcr = rgb2ycbcr(imgL);
    imgLY  = imgLYcbcr(:, :, 1);
    imgLCb = imgLYcbcr(:, :, 2);
    imgLCr = imgLYcbcr(:, :, 3);    
elseif 2 == ndims(imgL)
    imgLY = imgL;
else
    error('Input image error!');
end

% bicubic interpolation of the LR image
img = imresize(double(imgLY), scale, 'bicubic');
imgHY = zeros(size(img));
overlapMat = zeros(size(img));
[h, w] = size(img);

% extract LR image features
hf1 = [-1, 0, 1];
vf1 = [-1, 0, 1]';
hf2 = [1, 0, -2, 0, 1];
vf2 = [1, 0, -2, 0, 1]';
gradMapX1 = conv2(img, hf1, 'same');
gradMapY1 = conv2(img, vf1, 'same');
gradMapX2 = conv2(img, hf2, 'same');
gradMapY2 = conv2(img, vf2, 'same');
gradMag = sqrt(gradMapX1.^2 + gradMapY1.^2);

%% generate patch indexes
border = 0;
gridX = (border + 1) : patchSize - overlap : w - patchSize - border;
gridX = [gridX, w - patchSize - border];
gridY = (border + 1) : patchSize - overlap : h - patchSize - border;
gridY = [gridY, h - patchSize - border];

[X, Y] = meshgrid(gridX, gridY);
X = X(:); Y = Y(:);
numPatch = length(X);

%% build kd-tree for nearest neighbor search
metricTree = vl_kdtreebuild(C);
featTree = cell(numMetrics, 1);
featData = cell(numMetrics, 1);

for k = 1 : numMetrics
    featData{k} = M{k}*dictL(:, idxM == k);
    featTree{k} = vl_kdtreebuild(featData{k});
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% main loop to recover each LR patch
for i = 1 : numPatch        
    x = X(i); y = Y(i);
    
    % check gradient
    tempGradM = gradMag(y : y + patchSize - 1, x : x + patchSize -1);    
    if median(tempGradM(:)) <= gradThresh
        continue;
    end
    
    % LR patch
    imgPatch = img(y : y + patchSize - 1, x : x + patchSize -1);
    imgPatch = imgPatch(:);
    meanPatch = mean(imgPatch); 
    imgPatch = imgPatch - meanPatch;
    normPatch = norm(imgPatch, 1);    
    
    % LR feature   
    tempFeatX1 = gradMapX1(y : y + patchSize - 1, x : x + patchSize - 1);
    tempFeatY1 = gradMapY1(y : y + patchSize - 1, x : x + patchSize - 1);
    tempFeatX2 = gradMapX2(y : y + patchSize - 1, x : x + patchSize - 1);
    tempFeatY2 = gradMapY2(y : y + patchSize - 1, x : x + patchSize - 1);
    tempFeatX1 = tempFeatX1(:); tempFeatY1 = tempFeatY1(:);
    tempFeatX2 = tempFeatX2(:); tempFeatY2 = tempFeatY2(:);
    
    featPatch = [tempFeatX1./(norm(tempFeatX1, 1) + eps); ...
                 tempFeatY1./(norm(tempFeatY1, 1) + eps); ...
                 tempFeatX2./(norm(tempFeatX2, 1) + eps); ...
                 tempFeatY2./(norm(tempFeatY2, 1) + eps)];              
    
    % find nearest center
    [idxMetric, ~] = vl_kdtreequery(metricTree, C, featPatch, ...
        'NUMNEIGHBORS', 1, 'MAXNUMCOMPARISONS', 15);

    % find neighbours for regression
    A = M{idxMetric};
    idxDict = find(idxM == idxMetric);
    [idxNN, distNN] = vl_kdtreequery(featTree{idxMetric}, featData{idxMetric}, ...
        A*featPatch, 'NUMNEIGHBORS', numNN, 'MAXNUMCOMPARISONS', numNN*10);
    
    K = exp(-sqrt(distNN));
    K = K./(sum(K) + eps);
    
    % generate the HR patch and scale the contrast
    hPatch = dictH(:, idxDict(idxNN))*K;    
    hPatch = hPatch*normPatch;    
    hPatch = hPatch + meanPatch;
    hPatch = reshape(hPatch, [patchSize, patchSize]);
    
    imgHY(y : y + patchSize - 1, x : x + patchSize - 1) = ...
        imgHY(y : y + patchSize - 1, x : x + patchSize - 1) + hPatch;
    overlapMat(y : y + patchSize - 1, x : x + patchSize - 1) = ...
        overlapMat(y : y + patchSize - 1, x : x + patchSize - 1) + 1;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% fill in the empty with bicubic interpolation
idx = (overlapMat < 1);
imgHY(idx) = img(idx);
overlapMat(idx) = 1;
imgHY = imgHY./overlapMat;
imgHY = uint8(imgHY);

%% back projection
maxIter = 5;
imgHYBP = backProjection(imgHY, imgLY, maxIter);

%% combine color channels
if 3 == ndims(imgL)                
    imgHCb = imresize(imgLCb, scale, 'bicubic');
    imgHCr = imresize(imgLCr, scale, 'bicubic');
    imgH = zeros([h, w, 3]);
    imgH(:, :, 1) = imgHYBP;
    imgH(:, :, 2) = imgHCb;
    imgH(:, :, 3) = imgHCr;
    imgH = ycbcr2rgb(uint8(imgH)); 
else
    imgH = imgHYBP;
end

