function [dicH, dicL] = collectPatch(imgPath, gradThresh, format, ...
    scale, pSize, numPatch)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Collect Patches with Gradient Prune
%
% Input 
%    imgPath : file path for images
% gradThresh : threshold of gradient magnitude 
%     format : format of input image
%      scale : magnification factor
%      pSize : size of patch in low resolution (e.g. 5)
%   numPatch : number of patches to be collected per image
%
% Output
%       dicH : (dH x N) dictionary of HR patch
%       dicL : (dL x N) dictionary of LR patch
%
% Author
%   Renjie Liao
%   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

files = dir([imgPath '\*' format]);
numImgs = length(files);

dimPatch = pSize^2;
dicL = zeros(4*dimPatch, numImgs*numPatch);
dicH = zeros(dimPatch, numImgs*numPatch);

hf1 = [-1, 0, 1];
vf1 = [-1, 0, 1]';
hf2 = [1, 0, -2, 0, 1];
vf2 = [1, 0, -2, 0, 1]';    
idxDel = [];
border = 10;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for i = 1 : numImgs
    %% read image
    img = imread([imgPath '\' files(i, 1).name]);
        
    if 3 == ndims(img)
        imgYcbcr = rgb2ycbcr(img);
        imgH = double(imgYcbcr(:, :, 1));
    elseif 2 == ndims(img)
        imgH = double(img);
    else
        error('Input image error!');
    end
    
    [height, width] = size(imgH);
    imgL = imresize(imgH, 1/scale);
    imgL = imresize(imgL, [height, width]);    
    
    %% calculate feature
    % compute 1st and 2nd order gradients      
    gradMapX1 = conv2(imgL, hf1, 'same');
    gradMapY1 = conv2(imgL, vf1, 'same');
    gradMapX2 = conv2(imgL, hf2, 'same');
    gradMapY2 = conv2(imgL, vf2, 'same');    

    %% sampling with gradient prune
    gradMag = sqrt(gradMapX1.^2 + gradMapY1.^2);
    
    x = border + 1 : width - border - pSize + 1;
    y = border + 1 : height - border - pSize + 1;
    [X, Y] = meshgrid(x, y);
    X = X(:); Y = Y(:);
    numAllPatch = length(X);
    
    rng('shuffle');
    idx = randperm(numAllPatch);    
    count = 1;
    idxCount = 1;
    
    while idxCount <= numAllPatch                
        idxCol = X(idx(idxCount));
        idxRow = Y(idx(idxCount));
        tempGradM = gradMag(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);
        
        %% if gradient is small, discard it
        if median(tempGradM(:)) <= gradThresh
            idxCount = idxCount + 1;
            continue;
        end
        
        %% LR patch
        tempPatch = imgL(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);                
        tempPatch = tempPatch(:);
        meanPatch = mean(tempPatch);
        tempPatch = tempPatch - meanPatch;
        normPatch = norm(tempPatch, 1);
        
        %% LR feature
        tempFeatX1 = gradMapX1(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);
        tempFeatY1 = gradMapY1(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);
        tempFeatX2 = gradMapX2(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);
        tempFeatY2 = gradMapY2(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);
        
        tempFeatX1 = tempFeatX1(:);
        tempFeatY1 = tempFeatY1(:);
        tempFeatX2 = tempFeatX2(:);
        tempFeatY2 = tempFeatY2(:);        
        tempL = [tempFeatX1./(norm(tempFeatX1, 1) + eps); ...
                 tempFeatY1./(norm(tempFeatY1, 1) + eps); ...
                 tempFeatX2./(norm(tempFeatX2, 1) + eps); ...
                 tempFeatY2./(norm(tempFeatY2, 1) + eps)];                 
        
        %% HR patch
        tempH = imgH(idxRow : idxRow + pSize - 1, ...
            idxCol : idxCol + pSize - 1);
        tempH = tempH(:);
        tempH = tempH - meanPatch;
        tempH = tempH./(normPatch + eps);
        
        dicL(:, (i - 1)*numPatch + count) = tempL;
        dicH(:, (i - 1)*numPatch + count) = tempH;    
        
        count = count + 1; 
        idxCount = idxCount + 1;
        if count > numPatch
            break;
        end        
    end
    
    if count <= numPatch
        idxDel = [idxDel (i - 1)*numPatch + count : i*numPatch];
    end
    
    disp(['Finish ' num2str(i, '%03d') 'th image!']);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% delete empty entries
dicL(:, idxDel) = [];
dicH(:, idxDel) = [];
