function [M, C, idxM] = trainLLKR(dictH, dictL, numM, numNN, para)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Training Phase of Local Learnable Kernel Regression (LLKR) 
%
% Input
%   dictH : (dH x N) dictionary of HR patch
%   dictL : (dL x N) dictionary of LR patch
%    numM : number of local metrics  
%   numNN : number of nearest neighbors for LLKR
%    para : sturcture of parameters
% 
% Output
%    idxM : index of clustering
%       M : (k x 1 cell array) k metric matrix A
%       C : (dH x k) clustering centers
%
% Author
%   Renjie Liao
%   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    M = cell(numM, 1);
    dimL = size(dictL, 1);
    dimH = size(dictH, 1);
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% clustering
    [C, idxM] = vl_kmeans(dictL, numM, 'algorithm', 'elkan');
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% setting parameters
    startA = para.startA;
    stepSize = para.stepSize;
    maxIter = para.maxIter;
    epsilon = para.epsilon;
    numBatch = para.numBatch;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%    
    %% learning local metrics
    for k = 1 : numM
        A = diag(ones(dimL, 1)).*startA;
        idxDict = find(idxM == k);
        x = dictL(:, idxDict);
        y = dictH(:, idxDict);
        numSample = length(idxDict);
        yHatAll = zeros(dimH, numSample);
        gradScale = numSample/numBatch;
        
        errorFuncLast = .0;
        errorStart = .0;
        errorMin = inf;
        AStar = A;
        iter = 1;
                
        while iter <= maxIter                        
            %% update kd-tree
            dataX = A*x;
            myTree = vl_kdtreebuild(dataX);          
            
            %% calculate error function
            [idxNN, dist] = searchNN(myTree, dataX, dataX, numNN);
            K = exp(-sqrt(dist));
            K = bsxfun(@times, K, 1./(sum(K) + eps));
            
            for i = 1 : numSample
                yHatAll(:, i) = y(:, idxNN(:, i))*K(:, i);
            end
            
            errorFuncCur = sum((y(:) - yHatAll(:)).^2)*.5;
            
            if mod(iter, 100) == 1
                disp(['Error function of ' num2str(k, '%03d') ...
                      'th cluster || ' num2str(iter, '%03d') ...
                      'th iteration ' num2str(errorFuncCur, '%05f')]);
            end
            
            %% check stopping criterion
            if iter == 1
                errorStart = errorFuncCur;
            end
            
            if errorFuncCur < errorMin
                errorMin = errorFuncCur;
                AStar = A;
            end
            
            if abs(errorFuncCur - errorFuncLast)/errorFuncLast < epsilon                
                break;
            end
                
            %% Stochastic Gradient Descent
            % random sample patch 
            idxRand = randperm(numSample);
            if numSample < numBatch
                numBatch = numSample;
            end
            idx = idxRand(1:numBatch);

            % search nearest neighbors
            queryX = A*x(:, idx);            
            [idxNN1, ~] = searchNN(myTree, queryX, dataX, numNN);
            
            % calculate gradients
            gradA = zeros(dimL);
            for i = 1 : numBatch
                [yHat, K] = evalKR(queryX(:, i), dataX(:, idxNN1(:, i)), ...
                    y(:, idxNN1(:, i)));
                diffY = y(:, idx(i)) - yHat;
                [idxNN2, ~] = searchNN(myTree, dataX(:, idxNN1(:, i)), dataX, numNN);

                for j = 1 : numNN
                    diffX = dataX(:, idx(i)) - dataX(:, idxNN1(j, i));                    
                    [yHatNN, ~] = evalKR(dataX(:, idxNN1(j, i)), ...
                        dataX(:, idxNN2(:, j)), y(:, idxNN2(:, j)));
                    diffYHat = y(:, idxNN1(j, i)) - yHatNN;
                    
                    scalar = 2*(diffYHat'*diffY)*K(j);
                    gradA = gradA + diffX*diffX'.*scalar;
                end          
            end
            
            gradA = A*gradA;
            gradA = gradA.*gradScale;
            alpha = stepSize/sqrt(iter);            
            errorFuncLast = errorFuncCur;
            iter = iter + 1;
            
            A = A - gradA*alpha;            
        end
        
        disp(['The minimum error function value: ' num2str(errorMin, '%04f')...
            ' || Initial error function value: ' num2str(errorStart, '%04f')]);
        M{k} = AStar;
    end
end

%%
function [yq, K] = evalKR(xq, x, y)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Evaluate Kernel Regression 
%
% xq : (dL x 1) query data
%  x : (dL x N) retrieved data
%  y : (dH x N) target data
%  K : ( N x 1) kernel weight
% yq : (dH x 1) target of query
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	dx = bsxfun(@minus, xq, x);
	K = exp(-sqrt(sum(dx.^2))');
    K = K./(sum(K) + eps);    
	yq = y*K;
end

%%
function [idxNN, dist] = searchNN(tree, xq, x, numNN)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Search Nearest Neighbours 
%
%  tree : kd-tree structure
%    xq : (dL x M) query data
%     x : (dL x N) retrieved data
% numNN : number of nearest neighbors required
% idxNN : ( K x M) indexes of NNs
%  dist : ( K x M) distance between query and NNs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

    [idxNN, dist] = vl_kdtreequery(tree, x, xq, ...
            'NUMNEIGHBORS', numNN + 1, 'MAXNUMCOMPARISONS', 10*numNN);
    
    % exclude duplicate ones
    dist = dist(2:end, :);
    idxNN = double(idxNN(2:end, :));    
end
